"""Module contains Resource model."""

from typing import (  # noqa: F401 pylint: disable=unused-import
    Any,
    Dict,
    List,
    Tuple,
    Union,
)

from requests.models import Response  # noqa: F401 pylint: disable=unused-import

from crux.compat import unicode
from crux.models.label import Label
from crux.models.model import CruxModel
from crux.models.permission import Permission
from crux.utils import DEFAULT_CHUNK_SIZE, valid_chunk_size


class Resource(CruxModel):
    """Resource Model."""

    def __init__(
        self,
        id=None,  # type: str # id name is by design pylint: disable=redefined-builtin
        dataset_id=None,  # type: str
        folder_id=None,  # type: str
        name=None,  # type: str
        size=None,  # type: str
        type=None,  # type: str # type name is by design pylint: disable=redefined-builtin
        config=None,  # type: Dict[str, Any]
        provenance=None,  # type: str
        as_of=None,  # type: str
        created_at=None,  # type: str
        modified_at=None,  # type: str
        storage_id=None,  # type: str
        description=None,  # type: str
        media_type=None,  # type: str
        tags=None,  # type: List[str]
        labels=(),  # type: Tuple[Label,...]
        connection=None,
        raw_response=None,  # type: Dict[Any, Any]
    ):
        # type: (...) -> None
        """
        Attributes:
            id (str): Resourece Id. Defaults to None.
            dataset_id (str): Dataset Identity. Defaults to None.
            folder_id (str): Folder Identity. Defaults to None.
            name (str): Resource name. Defaults to None.
            size (str): Resource size. Defaults to None.
            type (str): Resource type. Defaults to None.
            config (str): Resource config. Defaults to None.
            provenance (list): Resource Provenance. Defaults to None.
            as_of (str): Resource as_of. Defaults to None.
            created_at (str): Resource created_at. Defaults to None.
            modified_at (str): Resource modified_at. Defaults to None.
            storage_id (str): Resource storage Identity. Defaults to None.
            description (str): Resource description. Defaults to None.
            media_type (str): Resource Media Type. Defaults to None.
            tags (:obj:`list` of :obj:`str`): Resource tags. Defaults to None.
            labels (:obj:`list` of :obj:`crux.models.Label`): List of Label objects.
                Defaults to None.
            connection (crux.client.CruxClient): Connection Object. Defaults to None.
            raw_response (dict): Response Content. Defaults to None.

        Raises:
            ValueError: If name or tags are set to None.
            TypeError: If tags are not of list type.
        """
        self._id = None
        self._dataset_id = None
        self._folder_id = None
        self._description = None
        self._name = None
        self._size = None
        self._type = None
        self._config = None
        self._provenance = None
        self._as_of = None
        self._created_at = None
        self._storage_id = None
        self._media_type = None
        self._modified_at = None
        self._tags = None
        self._folder = None
        self._labels = labels  # type: Tuple[Label, ...]

        self.id = id
        self.dataset_id = dataset_id
        self.folder_id = folder_id
        self.name = name
        self.size = size
        self.type = type
        self.config = config
        self.provenance = provenance
        self.as_of = as_of
        self.created_at = created_at
        self.modified_at = modified_at
        self.storage_id = storage_id
        self.connection = connection
        self.raw_response = raw_response
        self.media_type = media_type
        self.description = description
        self.tags = tags

    @property
    def id(self):
        """str: Gets and Sets the Resource ID."""
        return self._id

    @id.setter
    def id(self, id):  # id name is by design pylint: disable=redefined-builtin
        self._id = id

    @property
    def description(self):
        """str: Gets and Sets the Resource Description."""
        return self._description

    @description.setter
    def description(self, description):
        self._description = description

    @property
    def media_type(self):
        """str: Gets and Sets the Resource Description."""
        return self._media_type

    @media_type.setter
    def media_type(self, media_type):
        self._media_type = media_type

    @property
    def dataset_id(self):
        """str: Gets and Sets the Dataset ID."""
        return self._dataset_id

    @dataset_id.setter
    def dataset_id(self, dataset_id):
        self._dataset_id = dataset_id

    @property
    def folder_id(self):
        """str: Gets and Sets the Folder ID."""
        return self._folder_id

    @folder_id.setter
    def folder_id(self, folder_id):
        self._folder_id = folder_id

    @property
    def storage_id(self):
        """str: Gets and Sets the Storage ID."""
        return self._storage_id

    @storage_id.setter
    def storage_id(self, storage_id):
        self._storage_id = storage_id

    @property
    def name(self):
        """str: Gets and Sets the Resource Name."""
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")
        self._name = name

    @property
    def config(self):
        """str: Gets and Sets the config."""
        return self._config

    @config.setter
    def config(self, config):
        self._config = config

    @property
    def provenance(self):
        """str: Gets and Sets the Provenance."""
        return self._provenance

    @provenance.setter
    def provenance(self, provenance):
        self._provenance = provenance

    @property
    def type(self):
        """str: Gets and Sets the Resource Type."""
        return self._type

    @type.setter
    def type(self, type):  # type name is by design pylint: disable=redefined-builtin
        self._type = type

    @property
    def tags(self):
        """:obj:`list` of :obj:`str`: Gets and Sets the Resource Tags."""
        return self._tags

    @tags.setter
    def tags(self, tags):
        if not isinstance(tags, list):
            raise TypeError("Invalid Type for `tags`, tags should be of type list")
        self._tags = tags

    @property
    def labels(self):
        """:obj:`list` of :obj:`dict`: Gets and Sets the Resource labels."""
        return tuple(self._labels)

    @property
    def as_of(self):
        """str: Gets and Sets the as_of."""
        return self._as_of

    @as_of.setter
    def as_of(self, as_of):
        self._as_of = as_of

    @property
    def created_at(self):
        """str: Gets and Sets the created_at."""
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        self._created_at = created_at

    @property
    def modified_at(self):
        """str: Gets and Sets the modified_at."""
        return self._modified_at

    @modified_at.setter
    def modified_at(self, modified_at):
        self._modified_at = modified_at

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Resource object to Resource Dictionary.

        Returns:
            dict: Resource Dictionary.
        """
        return {
            "resourceId": self.id,
            "datasetId": self.dataset_id,
            "description": self.description,
            "folderId": self.folder_id,
            "mediaType": self.media_type,
            "name": self.name,
            "size": self.size,
            "type": self.type,
            # "config": self.config,
            # "provenance": self.provenance,
            "asof": self.as_of,
            "tags": self.tags,
            "labels": self.labels,
            "storageId": self.storage_id,
            "createdAt": self.created_at,
            "modifiedAt": self.modified_at,
        }

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, Any]) -> Any
        """Transforms Resource Dictionary to Resource object.

        Args:
            a_dict (dict): Resource Dictionary.

        Returns:
            crux.models.Resource: Resource Object.
        """
        id = a_dict[  # id name is by design pylint: disable=redefined-builtin
            "resourceId"
        ]
        dataset_id = a_dict["datasetId"]
        description = a_dict["description"]
        folder_id = a_dict["folderId"]
        storage_id = a_dict["storageId"]
        media_type = a_dict["mediaType"]
        description = a_dict["description"]
        name = a_dict["name"]
        # Added to provide compatibility with normal objects and Stitching objects,
        # till the time it is resolved
        if "tags" in a_dict:
            tags = a_dict["tags"]
        else:
            tags = []
        type = a_dict[  # type name is by design pylint: disable=redefined-builtin
            "type"
        ]
        # Added to provide compatibility with normal objects and Stitching objects,
        # till the time it is resolved
        if "config" in a_dict:
            config = a_dict["config"]
        else:
            config = None
        if "labels" in a_dict:
            labels = []
            for label in a_dict["labels"]:
                labels.append(Label.from_dict(label))
        else:
            labels = []
        provenance = a_dict["provenance"]
        created_at = a_dict["createdAt"]
        modified_at = a_dict["modifiedAt"]

        return cls(
            dataset_id=dataset_id,
            id=id,
            folder_id=folder_id,
            media_type=media_type,
            storage_id=storage_id,
            description=description,
            name=name,
            tags=tags,
            labels=tuple(labels),
            type=type,
            config=config,
            provenance=provenance,
            created_at=created_at,
            modified_at=modified_at,
        )

    def delete(self):
        # type: () -> bool
        """Deletes Resource from Dataset.

        Returns:
            bool: True if it is deleted.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        return self.connection.api_call(
            "DELETE", ["resources", self.id], headers=headers
        )

    def update(self, name=None, description=None, tags=None):
        # type: (str, str, List[str]) -> bool
        """Updates the metadata for Resource.

        Args:
            name (str): Name of resource. Defaults to None.
            description (str): Description of the resource. Defaults to None.
            tags (:obj:`list` of :obj:`str`): List of tags. Default to None.

        Returns:
            bool: True, if resource is updated.

        Raises:
            ValueError: It is raised if name, description or tags are unset.
            TypeError: It is raised if tags are not of type List.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        params = {}  # type: Dict[str, Union[str, List, Dict]]

        if name is not None:
            params["name"] = name
        if description is not None:
            params["description"] = description
        if tags is not None:
            if isinstance(tags, list):
                params["tags"] = tags
            else:
                raise TypeError("Tags should be of type list")

        if params:
            response = self.connection.api_call(
                "PUT", ["resources", self.id], headers=headers, params=params
            )

            response_dict = response.json()

            if "name" in response_dict:
                self.name = response.json().get("name")
            if "response" in response_dict:
                self.tags = response.json().get("tags")
            if "description" in response_dict:
                self.description = response.json().get("description")
            return True
        else:
            raise ValueError("Name, Description or Tags should be set")

    def _download(self, local_path, content_type, chunk_size=DEFAULT_CHUNK_SIZE):
        if content_type is not None:
            headers = {"Accept": content_type}
        else:
            headers = None

        if not valid_chunk_size(chunk_size):
            raise ValueError("chunk_size should be multiple of 256 KiB")

        data = self.connection.api_call(
            "GET", ["resources", self.id, "content"], headers=headers, stream=True
        )

        if hasattr(local_path, "write"):
            for chunk in data.iter_content(chunk_size=chunk_size):
                local_path.write(chunk)
            local_path.flush()
            return True
        elif isinstance(local_path, (str, unicode)):
            with open(local_path, mode="wb") as local_file:
                for chunk in data.iter_content(chunk_size=chunk_size):
                    local_file.write(chunk)
            return True
        else:
            raise TypeError(
                "Invalid Data Type for local_path: {}".format(type(local_path))
            )

    def download(self, local_path, content_type=None, chunk_size=DEFAULT_CHUNK_SIZE):
        # type: (...) -> bool
        """Downloads the resource.

        Args:
            local_path (str or file): Local OS path at which resource will be downloaded.
            content_type (str): Content Type of resource to be downloaded.
                Defaults to None.
            chunk_size (int): Number of bytes to be read in memory.

        Returns:
            bool: True if it is downloaded.

        Raises:
            ValueError: If chunk size is not multiple of 256 KB.
            TypeError: If local_path is not str or file object.
        """
        if not valid_chunk_size(chunk_size):
            raise ValueError("chunk_size should be multiple of 256 KiB")

        return self._download(
            local_path=local_path, content_type=content_type, chunk_size=chunk_size
        )

    def add_permission(self, identity_id="_subscribed_", permission="Read"):
        # type: (str, str) -> Union[bool, Permission]
        """Adds permission to the resource.

        Args:
            identity_id: Identity Id to be set. Defaults to _subscribed_.
            permission: Permission to be set. Defaults to Read.

        Returns:
            crux.models.Permission: Permission Object.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        return self.connection.api_call(
            "PUT",
            ["permissions", self.id, identity_id, permission],
            model=Permission,
            headers=headers,
        )

    def delete_permission(self, identity_id="_subscribed_", permission="Read"):
        # type: (str, str) -> bool
        """Deletes permission from the resource.

        Args:
            identity_id (str): Identity Id for the deletion.
                Defaults to _subscribed_.
            permission (str): Permission for the deletion.
                Defaults to Read.

        Returns:
            bool: True if it is able to delete it.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        return self.connection.api_call(
            "DELETE", ["permissions", self.id, identity_id, permission], headers=headers
        )

    def list_permissions(self):
        # type: () -> List[Permission]
        """Lists the permission on the resource.

        Returns:
            list (:obj:`crux.models.Permission`): List of Permission Objects.
        """
        headers = {"Accept": "application/json"}
        return self.connection.api_call(
            "GET",
            ["resources", self.id, "permissions"],
            model=Permission,
            headers=headers,
        )

    def add_label(self, label_key, label_value):
        # type: (str, str) -> bool
        """Adds label to Resource.

        Args:
            label_key (str): Label Key for Resource.
            label_value (str): Label Value for Resource.

        Returns:
            bool: True if label is added, False otherwise.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        response_result = self.connection.api_call(
            "PUT",
            [
                "datasets",
                self.dataset_id,
                "resources",
                self.id,
                "labels",
                label_key,
                label_value,
            ],
            headers=headers,
        )

        if response_result:
            self._labels = self._labels + (Label(label_key, label_value),)
            return True
        else:
            return False

    def delete_label(self, label_key):
        # type: (str) -> bool
        """Deletes label from Resource.

        Args:
            label_key (str): Label Key for Resource.

        Returns:
            bool: True if label is deleted, False otherwise.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        response_result = self.connection.api_call(
            "DELETE",
            ["datasets", self.dataset_id, "resources", self.id, "labels", label_key],
            headers=headers,
        )

        if response_result:
            for index, label in enumerate(self.labels):
                if label.label_key == label_key:
                    labels_list = list(self._labels)  # type: List[Label]
                    labels_list.pop(index)
                    self._labels = tuple(labels_list)
            return True
        else:
            return False

    def get_label(self, label_key):
        # type: (str) -> Label
        """Gets label value of Resource.

        Args:
            label_key (str): Label Key for Resource.

        Returns:
            crux.models.Label: Label Object.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        return self.connection.api_call(
            "GET",
            ["datasets", self.dataset_id, "resources", self.id, "labels", label_key],
            headers=headers,
            model=Label,
        )

    def get_all_labels(self):
        # type: () -> Tuple[Label, ...]
        """Fetches all labels of the Resource.

        Returns:
            list (:obj:`crux.models.Label`): List of Label Objects.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        response = self.connection.api_call(
            "GET", ["resources", self.id, "labels"], headers=headers
        )

        label_objects = []  # type: List[Label]

        labels = response.json().get("labels")

        if labels:
            for label in labels:
                obj = Label.from_dict(label)
                label_objects.append(obj)

        label_tuple_obj = tuple(label_objects)  # type: Tuple[Label, ...]
        return label_tuple_obj



"""Module contains Table model."""

from typing import Any, Dict  # noqa: F401 pylint: disable=unused-import

from crux.models.resource import Resource


class Table(Resource):
    """Table model."""

    @property
    def folder(self):
        """str: Gets and Sets the folder."""
        return self._folder

    @folder.setter
    def folder(self, folder):
        self._folder = folder

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Table object to Table Dictionary.

        Returns:
            dict: Table Dictionary.
        """
        return {
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "folder": self.folder,
            "type": self.type,
            "config": self.config,
            "labels": self.labels,
        }



"""Module contains code pertaining to CruxClient."""

from typing import (  # noqa: F401 pylint: disable=unused-import
    Any,
    Dict,
    List,
    MutableMapping,
    Optional,
    Text,
    Tuple,
    Union,
)

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import (
    ConnectTimeout,
    HTTPError,
    ProxyError,
    ReadTimeout,
    SSLError,
    TooManyRedirects,
)
from requests.packages.urllib3.util.retry import (  # Dynamic load pylint: disable=import-error
    Retry,
)

from crux.config import CruxConfig
from crux.exceptions import (
    CruxAPIError,
    CruxClientConnectionError,
    CruxClientHTTPError,
    CruxClientTimeout,
    CruxResourceNotFoundError,
)
from crux.utils import url_builder


class CruxClient(object):
    """Crux HTTP REST client."""

    def __init__(self, crux_config):
        # type: (CruxConfig) -> None
        if crux_config is None:
            self.crux_config = CruxConfig()  # type: CruxConfig
        else:
            self.crux_config = crux_config  # type: CruxConfig

    def api_call(  # pylint: disable=too-many-branches,too-many-statements
        self,
        method,  # type: str
        path,  # type: List[str]
        model=None,  # type: Any
        headers=None,  # type: MutableMapping[Text, Text]
        params=None,  # type: Dict[Any,Any]
        data=None,  # type: Dict[Any,Any]
        stream=False,  # type: bool
        retries=20,  # type: int
        backoff=0.3,  # type: float
        status_forcelist=(
            500,
            502,
            503,
            504,
            520,
            521,
            522,
            523,
            524,
            525,
            527,
            530,
        ),  # type: Tuple
        retry_on_methods=("GET", "PUT", "DELETE", "POST"),  # type: Tuple
        max_http_redirects=10,  # type: int
        timeout=60,  # type: int
        max_conn_errors=10,  # type: int
    ):
        # type:(...) -> Any
        """
        Requests and Serializes response from API Backend.

        Args:
            method (str): REST method name.
            path (str): API resource path.
            model (crux.models.CruxModel): Deserialization Model. Defaults to None.
            headers (dict): Additonal header parameters. Defaults to None.
            params (dict): Body data to be passed with request. Defaults to None.
            data (dict): Should be used while passing form encoded data. Defaults to None.
            stream (bool): Should be set to True, when response is required to be streamed.
                Defaults to False.
            retries (int): Total Retries to be performed. Defaults to 20.
            backoff (float): Backoff factor to be applied. Defaults to 0.3.
                {backoff factor} * (2 ^ ({number of total retries} - 1))
            status_forcelist (tuple): A set of integer HTTP status codes
                that should be retried on.
                Defaults to (500, 502, 503, 504, 520, 521, 522, 523, 524, 525, 527, 530).
            retry_on_methods (tuple): A set of uppercased HTTP method verbs
                that we should retry on.
                Defaults to ("GET", "PUT", "DELETE", "POST").
            max_http_redirects (int): Max HTTP sredirects to perform on API calls.
                Defaults to 10.
            timeout (int): Request timeout configuration in seconds.
                Defaults to 60.
            max_conn_errors (int): Max connection-related errors to retry on.
                Defaults to 10.

        Returns:
            crux.models.Model or bool: Serialized response from API backend.

        Raises:
            TypeError: If Path is not of list type.
            CruxClientHTTPError: If there is HTTP related error.
            CruxClientConnectionError: If there is SSL or Proxy related error.
            CruxClientTimeout: If there is timout related error.
            CruxResourceNotFoundError: If API has status code 400.
            CruxAPIError: If API has status code other than 2XX.
        """

        if path is None or not isinstance(path, list):
            raise TypeError("Path cannot be of NoneType. It should be of Type List")

        url = url_builder(
            url_base=self.crux_config.api_host,
            url_prefix=self.crux_config.api_prefix,
            url_path_list=path,
        )

        if headers is None:
            headers = {}

        if params is None:
            params = {}

        auth_scheme = "Bearer"  # type: str
        bearer_token = "{scheme} {key}".format(
            scheme=auth_scheme, key=self.crux_config.api_key
        )  # type: Text

        user_agent = self.crux_config.user_agent  # type: Text

        headers.update({"Authorization": bearer_token, "User-Agent": user_agent})

        retry = Retry(
            total=retries,
            backoff_factor=backoff,
            status_forcelist=status_forcelist,
            method_whitelist=retry_on_methods,
            redirect=max_http_redirects,
            connect=max_conn_errors,
        )

        adapter = HTTPAdapter(max_retries=retry)

        if method in ("GET", "DELETE"):
            try:
                with requests.session() as session:
                    session.mount("http://", adapter)
                    session.mount("https://", adapter)
                    response = session.request(
                        method,
                        url,
                        headers=headers,
                        params=params,
                        stream=stream,
                        proxies=self.crux_config.proxies,
                        timeout=timeout,
                    )
            except (HTTPError, TooManyRedirects) as err:
                raise CruxClientHTTPError(str(err))
            except (ProxyError, SSLError) as err:
                raise CruxClientConnectionError(str(err))
            except (ConnectTimeout, ReadTimeout) as err:
                raise CruxClientTimeout(str(err))
        elif method in ("PUT", "POST"):
            try:
                if data:
                    with requests.session() as session:
                        session.mount("http://", adapter)
                        session.mount("https://", adapter)
                        response = session.request(
                            method,
                            url,
                            headers=headers,
                            data=data,
                            proxies=self.crux_config.proxies,
                            timeout=timeout,
                        )
                else:
                    with requests.session() as session:
                        session.mount("http://", adapter)
                        session.mount("https://", adapter)
                        response = session.request(
                            method,
                            url,
                            headers=headers,
                            json=params,
                            proxies=self.crux_config.proxies,
                            timeout=timeout,
                        )
            except (HTTPError, TooManyRedirects) as err:
                raise CruxClientHTTPError(str(err))
            except (ProxyError, SSLError) as err:
                raise CruxClientConnectionError(str(err))
            except (ConnectTimeout, ReadTimeout) as err:
                raise CruxClientTimeout(str(err))
        else:
            raise ValueError("Request Method Type should be in GET, DELETE, PUT, POST")

        if response.status_code in (200, 201, 202, 206):
            if model is None:
                return response
            else:
                if isinstance(response.json(), list):
                    serial_list = []
                    for item in response.json():
                        obj = model.from_dict(item)
                        obj.connection = self
                        obj.raw_response = response.json()
                        serial_list.append(obj)
                    return serial_list
                else:
                    obj = model.from_dict(response.json())
                    obj.connection = self
                    obj.raw_response = response.json()
                    return obj
        elif response.status_code == 204:
            return True
        else:
            if response.status_code == 404:
                raise CruxResourceNotFoundError(response.json())
            else:
                raise CruxAPIError(response.json())



"""Modules contains set of utility functions."""

import os
import posixpath
from typing import List, Tuple  # noqa: F401 pylint: disable=unused-import

from crux.compat import Enum, urllib_quote


DEFAULT_CHUNK_SIZE = 10485760  # 10 MB


# https://github.com/python/mypy/issues/2477, mypy is performing checking with Python2
class ContentType(Enum):  # type: ignore
    """ContentType Enumeration Model."""

    JSON = "application/json"
    NDJSON = "application/x-ndjson"
    CSV = "text/csv"
    PARQUET = "application/parquet"
    AVRO = "avro/binary"

    @classmethod
    def detect(cls, file_name):
        # type: (str) -> str
        """Detects the content_type from the file extension.

        Args:
            file_name (str): Absolute or Relative Path of the file.

        Returns:
            str: ContentType extension.

        Raises:
            LookupError: If file type is not supported.
        """
        file_ext = os.path.splitext(file_name)[1][1:].upper()

        if file_ext in cls.__members__:
            return cls[file_ext].value  # type: ignore
        else:
            raise LookupError("File/Media Type not supported.")


def quote(data):
    # type: (str) -> str
    """Percent Encodes the string.

    Args:
        data (str): String which is required to be Percent Encoded.

    Returns:
        str: Percent encoded string.
    """

    # With current setting, it will encode all necessary characters.
    # To prevent any character from getting percent encoded, add it to safe_characters
    # For eg: safe_characters="/%$"
    # https://docs.python.org/2/library/urllib.html
    safe_characters = ""
    return urllib_quote(data, safe=safe_characters)


def url_builder(url_base, url_prefix, url_path_list):
    # type: (str, str, List[str]) -> str
    """Builds Valid Percent Encoded URL.

    Args:
        url_base (str): Base of URL comprised of HTTP protocol scheme and Hostname.
        url_prefix (str): URL Prefix used by API Platform.
        url_path_list (:obj:`list` of :obj:`str`): List of Context Path elements.

    Returns:
        str: Valid Percent Encoded URL.
    """
    if url_prefix:
        valid_url = [url_base, quote(url_prefix)]
    else:
        valid_url = [url_base]

    valid_url_path_list = []

    for component in url_path_list:
        valid_url_path_list.append(quote(component))

    valid_url.extend(valid_url_path_list)

    return "/".join(valid_url)


def valid_chunk_size(chunk_size):
    # type: (int) -> bool
    """Checks whether chunk size is multiple of 256 KiB.

    Args:
        chunk_size (int): Input chunk_size to be validated.

    Returns:
        bool: True if chunk_size is multiple of 256 KiB, False otherwise.
    """
    return not bool(chunk_size % 262144)  # 1024*256=262144


def split_posixpath_filename_dirpath(path):
    # type: (str) -> Tuple[str, str]
    """Split a POSIX path into file name and directory path.

    Args:
        path (str): POSIX path to split.

    Returns:
        (str, str): Returns a tuple of (file name, directory path) with
            trailing slash removed from directory.
    """

    filename = posixpath.basename(path)  # type: str
    dirpath = posixpath.dirname(path)  # type: str
    return filename, dirpath



"""Module provides CruxConfig object to manage API configuration settings."""

import os
import platform
import re
from typing import (  # noqa: F401 pylint: disable=unused-import
    Dict,
    MutableMapping,
    Optional,
    Text,
    Union,
)

import requests

from crux.__version__ import __version__


class CruxConfig(object):
    """
    Crux Configuration Class.
    """

    def __init__(
        self,
        api_key=None,  # type: Optional[str]
        api_host=None,  # type: str
        api_prefix=None,  # type: str
        proxies=None,  # type: Optional[MutableMapping[Text, Text]]
        user_agent=None,  # type: str
    ):
        # type: (...) -> None
        """
        Args:
            api_key (str): API Key. Defaults to None.
            api_host (str): API URL. Defaults to None.
            api_prefix (str): API prefix to be used. Defaults to None.
            proxies (dict): Proxies to be used. Defaults to None.
            user_agent (str): User agent to be used. Defaults to None.

        Raises:
            ValueError: If CRUX_AP_KEY is not set.
        """
        self.re_user_agent_banned_chars = re.compile(r"[^a-zA-Z0-9._+~-]")
        self.re_whitespace_runs = re.compile(r"\s+")

        if api_key is None:
            if "CRUX_API_KEY" in os.environ:
                self.api_key = os.environ.get("CRUX_API_KEY")  # type: Optional[str]
            else:
                raise ValueError("API KEY is required")
        else:
            self.api_key = api_key  # type: Optional[str]

        if api_host is None:
            self.api_host = os.environ.get(
                "CRUX_API_HOST", "https://api.cruxinformatics.com"
            )
        else:
            self.api_host = api_host

        if api_prefix is None:
            self.api_prefix = os.environ.get("CRUX_API_PREFIX", "plat-api")
        else:
            self.api_prefix = api_prefix

        if user_agent is None:
            self.user_agent = self._default_user_agent()
        else:
            self.user_agent = user_agent

        self.proxies = (
            proxies if proxies else {}
        )  # type: Optional[MutableMapping[Text, Text]]

    def _default_user_agent(self):
        # type: () -> str
        user_agent = (
            "crux-python/{ver}"
            " requests/{req_ver} {py_impl}/{py_ver} "
            "{os}/{os_ver} {cpu}/{machine}"
        ).format(
            ver=__version__,
            req_ver=self._sanitize_user_agent_part(requests.__version__),
            py_impl=self._sanitize_user_agent_part(platform.python_implementation()),
            py_ver=self._sanitize_user_agent_part(platform.python_version()),
            os=self._sanitize_user_agent_part(platform.system()),
            os_ver=self._sanitize_user_agent_part(platform.release()),
            cpu=self._sanitize_user_agent_part(platform.processor()),
            machine=self._sanitize_user_agent_part(platform.machine()),
        )

        return user_agent

    def _sanitize_user_agent_part(self, part):
        # type: (str) -> str
        if part:
            no_space_part = self.re_whitespace_runs.sub("_", part)
            sanitized_part = self.re_user_agent_banned_chars.sub("", no_space_part)
            if sanitized_part:
                return sanitized_part
        return "unknown"



"""Module contains File model."""

from typing import Any, Dict, Union  # noqa: F401 pylint: disable=unused-import

from crux.models.permission import Permission
from crux.models.resource import Resource


class Folder(Resource):
    """Folder Model."""

    @property
    def folder(self):
        """str: Gets and Sets the folder."""
        return self._folder

    @folder.setter
    def folder(self, folder):
        self._folder = folder

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Folder object to Folder Dictionary.

        Returns:
            dict: Folder Dictionary.
        """
        return {
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "folder": self.folder,
            "type": self.type,
            "labels": self.labels,
        }

    def add_permission(  # It is by design pylint: disable=arguments-differ
        self, identity_id="_subscribed_", permission="Read", recursive=False
    ):
        # type: (str, str, bool) -> Union[bool, Permission]
        """Adds permission to the Folder resource.

        Args:
            identity_id (str): Identity Id to be set. Defaults to _subscribed_.
            permission (str): Permission to be set. Defaults to Read.
            recursive (bool): If recursive is set to True, it will recursive apply
                permission to all resources under the folder resource.

        Returns:
            bool or crux.models.Permission: If recursive is set then it returns True.
                If recursive is unset then it returns Permission object.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        params = {
            "identityId": identity_id,
            "permission": permission,
            "action": "add",
            "resourceIds": [self.id],
        }
        if recursive:
            return self.connection.api_call(
                "POST", ["permissions", "bulk"], headers=headers, params=params
            )
        else:
            return self.connection.api_call(
                "PUT",
                ["permissions", self.id, identity_id, permission],
                model=Permission,
                headers=headers,
            )

    def delete_permission(  # It is by design pylint: disable=arguments-differ
        self, identity_id="_subscribed_", permission="Read", recursive=False
    ):
        # type: (str, str, bool) -> bool
        """Deletes permission from Folder resource.

        Args:
            identity_id (str): Identity Id for the deletion. Defaults to _subscribed_.
            permission (str): Permission for deletion. Defaults to Read.
            recursive (bool): If recursive is set to True, it will recursively delete
                permission from all resources under the folder resource.
                Defaults to False.

        Returns:
            bool: True if it is able to delete it.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        params = {
            "identityId": identity_id,
            "permission": permission,
            "action": "delete",
            "resourceIds": [self.id],
        }
        if recursive:
            return self.connection.api_call(
                "POST", ["permissions", "bulk"], headers=headers, params=params
            )
        else:
            return self.connection.api_call(
                "DELETE",
                ["permissions", self.id, identity_id, permission],
                headers=headers,
            )



"""Module contains File model."""

from typing import (  # noqa: F401 pylint: disable=unused-import
    Any,
    Dict,
    IO,
    Iterable,
    List,
    Union,
)

from crux.models.resource import Resource
from crux.utils import ContentType, DEFAULT_CHUNK_SIZE, valid_chunk_size


class File(Resource):
    """File Model."""

    @property
    def folder(self):
        """str: Gets and Sets the folder."""
        return self._folder

    @folder.setter
    def folder(self, folder):
        self._folder = folder

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms File object to File Dictionary.

        Returns:
            dict: File Dictionary.
        """
        return {
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "folder": self.folder,
            "type": self.type,
            "labels": self.labels,
        }

    def iter_content(self, chunk_size=DEFAULT_CHUNK_SIZE, decode_unicode=False):
        # type: (int, bool) -> Iterable[str]
        """Streams the file resource.

        Args:
            chunk_size (int): Chunk Size for the stream.
            decode_unicode (bool): If decode_unicode is True, content will be decoded using the
                best available encoding based on the response.

        Yields:
            bytes: Bytes of file resource.

        Raises:
            ValueError: If chunk_size is not multiple of 256 KiB.
        """
        headers = {"Accept": "*/*"}

        if not valid_chunk_size(chunk_size):
            raise ValueError("chunk_size should be multiple of 256 KiB")

        data = self.connection.api_call(
            "GET", ["resources", self.id, "content"], headers=headers, stream=True
        )

        return data.iter_content(chunk_size=chunk_size, decode_unicode=decode_unicode)

    def download(
        self, local_path, chunk_size=DEFAULT_CHUNK_SIZE
    ):  # content_type is skipped intentionally pylint: disable=arguments-differ
        # type: ignore # https://github.com/python/mypy/issues/3750
        """Downloads the file resource.

        Args:
            local_path (str or file): Local OS path at which file resource will be downloaded.
            chunk_size (int): Number of bytes to be read in memory.

        Returns:
            bool: True if it is downloaded.

        Raises:
            ValueError: If chunk_size is not multiple of 256 KiB.
        """
        if not valid_chunk_size(chunk_size):
            raise ValueError("chunk_size should be multiple of 256 KiB")

        return self._download(
            local_path=local_path, content_type=None, chunk_size=chunk_size
        )

    def upload(self, local_path, content_type=None):
        # type: (Union[IO, str], str) -> bool
        """Uploads the content to empty file resource.

        Args:
            local_path (str or file): Local OS path whose content is to be uploaded.
            content_type (str): Content type of the file. Defaults to None.

        Returns
            bool: True if it is uploaded.

        Raises:
            TypeError: If local_path type is invalid.
        """

        if hasattr(local_path, "read"):

            if content_type is None:
                content_type = ContentType.detect(getattr(local_path, "name"))

            headers = {"Content-Type": content_type, "Accept": "application/json"}

            resp = self.connection.api_call(
                "PUT",
                ["resources", self.id, "content"],
                data=local_path,
                headers=headers,
            )

            return resp.status_code == 200

        elif isinstance(local_path, str):

            if content_type is None:
                content_type = ContentType.detect(local_path)

            headers = {"Content-Type": content_type, "Accept": "application/json"}

            with open(local_path, mode="rb") as data:
                resp = self.connection.api_call(
                    "PUT", ["resources", self.id, "content"], data=data, headers=headers
                )

            return resp.status_code == 200

        else:
            raise TypeError("Invalid Data Type for local_path")
